# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VectorNetworkAnalyzer class."""

from abc import abstractmethod
from dataclasses import dataclass

from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.typings.enums import Parameter, VNAScatteringParameters, VNATriggerModes
from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver

DEFAULT_NUMBER_POINTS = 1000


class VectorNetworkAnalyzer(Instrument):
    """Abstract base class defining all vector network analyzers"""

    @dataclass
    class VectorNetworkAnalyzerSettings(Instrument.InstrumentSettings):
        """Contains the settings of a specific signal generator.

        Args:
            power (float): Power of the instrument in dBm
            scatering_parameter (str): Scatering parameter of the instrument
            frequency_span (float): Frequency span of the instrument in KHz
            frequency_center (float): Frequency center of the instrument in Hz
            frequency_start (float): Frequency start of the instrument in Hz
            frequency_stop (float): Frequency stop of the instrument in Hz
            if_bandwidth (float): Bandwidth of the instrument in Hz
            averaging_enabled (bool): Whether averaging is enabled or not in the instrument
            trigger_mode (str): Trigger mode of the instrument
            number_points (int): Number of points for sweep
            device_timeout (float): Timeout of the instrument in ms
            electrical_delay (float): Electrical delay of the instrument in s
        """

        power: float
        scattering_parameter: VNAScatteringParameters = VNAScatteringParameters.S11
        frequency_span: float | None = None
        frequency_center: float | None = None
        frequency_start: float | None = None
        frequency_stop: float | None = None
        if_bandwidth: float | None = None
        averaging_enabled: bool = False
        number_averages: int = 1
        trigger_mode: VNATriggerModes = VNATriggerModes.INT
        number_points: int = DEFAULT_NUMBER_POINTS
        electrical_delay: float = 0.0

    settings: VectorNetworkAnalyzerSettings
    device: VectorNetworkAnalyzerDriver

    def setup(self, parameter: Parameter, value: float | str | bool | int, channel_id: int | None = None):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (float | str | bool | int): new value
            channel_id (int | None): channel identifier of the parameter to update
        """
        if isinstance(value, str):
            self._set_parameter_str(parameter=parameter, value=value)
            return
        if isinstance(value, bool):
            self._set_parameter_bool(parameter=parameter, value=value)
            return
        if isinstance(value, float):
            self._set_parameter_float(parameter=parameter, value=value)
            return
        if isinstance(value, int):
            self._set_parameter_int(parameter=parameter, value=value)
            return
        raise ParameterNotFound(self, parameter)

    def _set_parameter_str(self, parameter: Parameter, value: str):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (str): new value
        """
        if parameter == Parameter.SCATTERING_PARAMETER:
            self.scattering_parameter = VNAScatteringParameters(value)
            return
        if parameter == Parameter.TRIGGER_MODE:
            self.settings.trigger_mode = VNATriggerModes(value)
            return

        raise ParameterNotFound(self, parameter)

    def _set_parameter_bool(self, parameter: Parameter, value: bool):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (bool): new value
        """
        if parameter == Parameter.AVERAGING_ENABLED:
            self.averaging_enabled = value
            return

        raise ParameterNotFound(self, parameter)

    def _set_parameter_float(self, parameter: Parameter, value: float) -> None:
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (float): new value
        """

        if parameter == Parameter.POWER:
            self.set_power(value)
            return
        if parameter == Parameter.FREQUENCY_SPAN:
            self.frequency_span = value
            return
        if parameter == Parameter.FREQUENCY_CENTER:
            self.frequency_center = value
            return
        if parameter == Parameter.FREQUENCY_START:
            self.frequency_start = value
            return
        if parameter == Parameter.FREQUENCY_STOP:
            self.frequency_stop = value
            return
        if parameter == Parameter.IF_BANDWIDTH:
            self.set_if_bandwidth(value)
            return
        if parameter == Parameter.ELECTRICAL_DELAY:
            self.electrical_delay = value
            return

        raise ParameterNotFound(self, parameter)

    def _set_parameter_int(self, parameter: Parameter, value: int):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (int): new value
        """

        if parameter == Parameter.NUMBER_AVERAGES:
            self.number_averages = value
            return

        if parameter == Parameter.NUMBER_POINTS:
            self.number_points = value
            return

        raise ParameterNotFound(self, parameter)

    def get_power(self):
        """VectorNetworkAnalyzer 'power' property.

        Returns:
            float: settings.power
        """
        return self.settings.power

    @abstractmethod
    def set_power(self, power: float, channel=1, port=1):
        """sets the power in dBm"""

    @property
    def scattering_parameter(self):
        """VectorNetworkAnalyzer 'scattering_parameter' property.

        Returns:
            str: settings.scattering_parameter.
        """
        return self.settings.scattering_parameter

    @scattering_parameter.setter
    def scattering_parameter(self, value: str, channel=1):
        """sets the scattering parameter"""
        self.settings.scattering_parameter = VNAScatteringParameters(value)

        if self.is_device_active():
            scat_par = self.settings.scattering_parameter.value
            self.send_command(f"CALC1:MEAS{channel}:PAR", scat_par)

    @property
    def frequency_span(self):
        """VectorNetworkAnalyzer 'frequency_span' property.

        Returns:
            float: settings.frequency_span.
        """
        return self.settings.frequency_span

    @frequency_span.setter
    def frequency_span(self, value: float, channel=1):
        """sets the frequency span in kHz"""
        self.settings.frequency_span = value

        if self.is_device_active():
            freq = str(self.settings.frequency_span)
            self.send_command(f"SENS{channel}:FREQ:SPAN", freq)

    @property
    def frequency_center(self):
        """VectorNetworkAnalyzer 'frequency_center' property.

        Returns:
            float: settings.frequency_center.
        """
        return self.settings.frequency_center

    @frequency_center.setter
    def frequency_center(self, value: float, channel=1):
        """sets the frequency center in Hz"""
        self.settings.frequency_center = value

        if self.is_device_active():
            freq = str(self.settings.frequency_center)
            self.send_command(f"SENS{channel}:FREQ:CENT", freq)

    @property
    def frequency_start(self):
        """VectorNetworkAnalyzer 'frequency_start' property.

        Returns:
            float: settings.frequency_start.
        """
        return self.settings.frequency_start

    @frequency_start.setter
    def frequency_start(self, value: float, channel=1):
        """sets the frequency start in Hz"""
        self.settings.frequency_start = value

        if self.is_device_active():
            freq = str(self.settings.frequency_start)
            self.send_command(f"SENS{channel}:FREQ:STAR", freq)

    @property
    def frequency_stop(self):
        """VectorNetworkAnalyzer 'frequency_stop' property.

        Returns:
            float: settings.frequency_stop.
        """
        return self.settings.frequency_stop

    @frequency_stop.setter
    def frequency_stop(self, value: float, channel=1):
        """sets the frequency stop in Hz"""
        self.settings.frequency_stop = value

        if self.is_device_active():
            freq = str(self.settings.frequency_stop)
            self.send_command(f"SENS{channel}:FREQ:STOP", freq)

    def get_if_bandwidth(self):
        """VectorNetworkAnalyzer 'if_bandwidth' property.

        Returns:
            float: settings.if_bandwidth.
        """
        return self.settings.if_bandwidth

    @abstractmethod
    def set_if_bandwidth(self, value: float, channel=1):
        """sets the if bandwidth in Hz"""

    @property
    def averaging_enabled(self):
        """VectorNetworkAnalyzer 'averaging_enabled' property.

        Returns:
            bool: settings.averaging_enabled.
        """
        return self.settings.averaging_enabled

    @averaging_enabled.setter
    def averaging_enabled(self, value: bool):
        """sets the averaging enabled"""
        self.settings.averaging_enabled = value

        if self.is_device_active():
            self._average_state(state=self.settings.averaging_enabled)

    @property
    def number_averages(self):
        """VectorNetworkAnalyzer 'number_averages' property.

        Returns:
            int: settings.number_averages.
        """
        return self.settings.number_averages

    @number_averages.setter
    def number_averages(self, value: int, channel=1):
        """sets the number averages"""
        self.settings.number_averages = value

        if self.is_device_active():
            self._average_count(count=str(self.settings.number_averages), channel=channel)

    @property
    def trigger_mode(self):
        """VectorNetworkAnalyzer 'trigger_mode' property.

        Returns:
            str: settings.trigger_mode.
        """
        return self.settings.trigger_mode

    @property
    def number_points(self):
        """VectorNetworkAnalyzer 'number_points' property.

        Returns:
            int: settings.number_points.
        """
        return self.settings.number_points

    @number_points.setter
    def number_points(self, value: int, channel=1):
        """sets the number of points for sweep"""
        self.settings.number_points = value

        if self.is_device_active():
            points = str(self.settings.number_points)
            self.send_command(f":SENS{channel}:SWE:POIN", points)

    @property
    def electrical_delay(self):
        """VectorNetworkAnalyzer 'electrical_delay' property.

        Returns:
            float: settings.electrical_delay.
        """
        return self.settings.electrical_delay

    @electrical_delay.setter
    @abstractmethod
    def electrical_delay(self, value: float):
        """
        Set electrical delay in channel 1

        Input:
            value (str) : Electrical delay in ns
                example: value = '100E-9' for 100ns
        """

    def _average_state(self, state, channel=1):
        """Set status of Average."""
        if state:
            self.send_command(f"SENS{channel}:AVER:STAT", "ON")
        else:
            self.send_command(f"SENS{channel}:AVER:STAT", "OFF")

    def _average_count(self, count, channel):
        """Set the average count"""
        self.send_command(f"SENS{channel}:AVER:COUN", count)
        self.send_command(command=f":SENS{channel}:AVER:CLE", arg="")

    def to_dict(self):
        """Return a dict representation of the VectorNetworkAnalyzer class."""
        return dict(super().to_dict().items())

    def initial_setup(self):
        """Set initial instrument settings."""
        self.device.initial_setup()

    def reset(self):
        """Reset instrument settings."""
        self.device.reset()

    def turn_on(self):
        """Start an instrument."""
        return self.send_command(command=":OUTP", arg="ON")

    def turn_off(self):
        """Stop an instrument."""
        return self.send_command(command=":OUTP", arg="OFF")

    def send_command(self, command: str, arg: str = "?") -> str:
        """Send a command directly to the device.

        Args:
            command(str): Command to send the device
            arg(str): Argument to send the command with. Default empty string

        Example:
            >>> send_command(command=":OUTP",arg="ON") -> ":OUTP ON"
        """
        return self.device.send_command(command=command, arg=arg)

    def send_query(self, query: str):
        """
        Send a query directly to the device.

        Input:
            query(str): Query to send the device
        """
        return self.device.send_query(query)

    def send_binary_query(self, query: str):
        """
        Send a binary query directly to the device.

        Input:
            query(str): Query to send the device
        """
        return self.device.send_binary_query(query)

    def autoscale(self):
        """Autoscale"""
        self.send_command(command="DISP:WIND:TRAC:Y:AUTO", arg="")

    def read(self) -> str:
        """Read directly from the device"""
        return self.device.read()

    def read_raw(self) -> str:
        """Read raw data directly from the device"""
        return self.device.read_raw()

    def set_timeout(self, value: float):
        """Set timeout in mili seconds"""
        self.device.set_timeout(value)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bus class."""

import contextlib
from dataclasses import InitVar, dataclass, field

from qpysequence import Sequence as QpySequence

from qililab.constants import RUNCARD
from qililab.instruments import Instrument, Instruments, ParameterNotFound
from qililab.instruments.qblox import QbloxQCM, QbloxQRM
from qililab.pulse.pulse_distortion.pulse_distortion import PulseDistortion
from qililab.qprogram.qblox_compiler import AcquisitionData
from qililab.result import Result
from qililab.result.qprogram import MeasurementResult
from qililab.settings import Settings
from qililab.typings import ChannelID, Parameter, ParameterValue


class Bus:
    """Bus class.

    Ideally a bus should contain a qubit control/readout and a signal generator, which are connected
    through a mixer for up- or down-conversion. At the end of the bus there should be a qubit or a resonator object,
    which is connected to one or multiple qubits.

    Args:
        settings (BusSettings): Bus settings.
    """

    @dataclass
    class BusSettings(Settings):
        """Bus settings.

        Args:
            alias (str): Alias of the bus.
            system_control (SystemControl): System control used to control and readout the qubits of the bus.
            port (str): Alias of the port where bus is connected.
            distortions (list[PulseDistotion]): List of the distortions to apply to the Bus.
            delay (int): Bus delay
        """

        alias: str
        instruments: list[Instrument]
        channels: list[ChannelID | None]
        platform_instruments: InitVar[Instruments]
        delay: int = 0
        distortions: list[PulseDistortion] = field(default_factory=list)

        def __post_init__(self, platform_instruments: Instruments):  # type: ignore
            instruments = []
            for inst_alias in self.instruments:
                inst_class = platform_instruments.get_instrument(alias=inst_alias)  # type: ignore
                if inst_class is None:
                    raise NameError(
                        f"The instrument with alias {inst_alias} could not be found within the instruments of the "
                        "platform. The available instrument aliases are: "
                        f"{[inst.alias for inst in platform_instruments.elements]}."
                    )
                instruments.append(inst_class)
            self.instruments = instruments
            self.distortions = [
                PulseDistortion.from_dict(distortion)  # type: ignore[arg-type]
                for distortion in self.distortions
                if isinstance(distortion, dict)
            ]
            super().__post_init__()

    settings: BusSettings
    """Bus settings. Containing the alias of the bus, the system control used to control and readout its qubits, the alias
    of the port where it's connected, the list of the distortions to apply, and its delay.
    """

    def __init__(self, settings: dict, platform_instruments: Instruments):
        self.settings = self.BusSettings(**settings, platform_instruments=platform_instruments)  # type: ignore[call-arg]

    @property
    def alias(self):
        """Alias of the bus.

        Returns:
            str: alias of the bus
        """
        return self.settings.alias

    @property
    def instruments(self) -> list[Instrument]:
        """Instruments controlled by this system control."""
        return self.settings.instruments

    @property
    def channels(self) -> list[ChannelID | None]:
        """Instruments controlled by this system control."""
        return self.settings.channels

    @property
    def delay(self):
        """Bus 'delay' property.

        Returns:
            int: settings.delay.
        """
        return self.settings.delay

    @property
    def distortions(self):
        """Bus 'distortions' property.

        Returns:
            list[PulseDistortion]: settings.distortions.
        """
        return self.settings.distortions

    def __str__(self):
        """String representation of a bus. Prints a drawing of the bus elements."""
        instruments = "--".join(f"|{instrument}|" for instrument in self.instruments)
        return f"Bus {self.alias}:  ----{instruments}----"

    def __eq__(self, other: object) -> bool:
        """compare two Bus objects"""
        return str(self) == str(other) if isinstance(other, Bus) else False

    def __iter__(self):
        """Redirect __iter__ magic method."""
        return iter(zip(self.instruments, self.channels))

    def to_dict(self):
        """Return a dict representation of the Bus class."""
        return {
            RUNCARD.ALIAS: self.alias,
            RUNCARD.INSTRUMENTS: [instrument.alias for instrument in self.instruments],
            "channels": self.settings.channels,
        }

    def has_awg(self) -> bool:
        """Return true if bus has AWG capabilities."""
        return any(instrument.is_awg() for instrument in self.instruments)

    def has_adc(self) -> bool:
        """Return true if bus has ADC capabilities."""
        return any(instrument.is_adc() for instrument in self.instruments)

    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set a parameter to the bus.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (int | float | str | bool): value to update
            channel_id (int, optional): instrument channel to update, if multiple. Defaults to None.
        """
        for instrument, instrument_channel in zip(self.instruments, self.channels):
            with contextlib.suppress(ParameterNotFound):
                if channel_id is not None and channel_id == instrument_channel:
                    instrument.set_parameter(parameter, value, channel_id)
                    return
                instrument.set_parameter(parameter, value, instrument_channel)
                return
        raise Exception(f"No parameter with name {parameter.value} was found in the bus with alias {self.alias}")

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Gets a parameter of the bus.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (int | float | str | bool): value to update
            channel_id (int, optional): instrument channel to update, if multiple. Defaults to None.
        """
        if parameter == Parameter.DELAY:
            return self.settings.delay
        for instrument, instrument_channel in zip(self.instruments, self.channels):
            with contextlib.suppress(ParameterNotFound):
                if channel_id is not None and channel_id == instrument_channel:
                    return instrument.get_parameter(parameter, channel_id)
                return instrument.get_parameter(parameter, instrument_channel)
        raise Exception(f"No parameter with name {parameter.value} was found in the bus with alias {self.alias}")

    def upload_qpysequence(self, qpysequence: QpySequence):
        """Uploads the qpysequence into the instrument."""
        from qililab.instruments.qblox.qblox_module import QbloxModule  # pylint: disable=import-outside-toplevel

        for instrument, instrument_channel in zip(self.instruments, self.channels):
            if isinstance(instrument, QbloxModule):
                instrument.upload_qpysequence(qpysequence=qpysequence, channel_id=int(instrument_channel))  # type: ignore[arg-type]
                return

        raise AttributeError(f"Bus {self.alias} doesn't have any QbloxModule to upload a qpysequence.")

    def upload(self):
        """Uploads any previously compiled program into the instrument."""
        for instrument, instrument_channel in zip(self.instruments, self.channels):
            if isinstance(instrument, (QbloxQCM, QbloxQRM)):
                instrument.upload(channel_id=instrument_channel)
                return

    def run(self) -> None:
        """Runs any previously uploaded program into the instrument."""
        for instrument, instrument_channel in zip(self.instruments, self.channels):
            if isinstance(instrument, (QbloxQCM, QbloxQRM)):
                instrument.run(channel_id=instrument_channel)  # type: ignore
                return

    def acquire_result(self) -> Result:
        """Read the result from the vector network analyzer instrument

        Returns:
            Result: Acquired result
        """
        # TODO: Support acquisition from multiple instruments
        results: list[Result] = []
        for instrument in self.instruments:
            if isinstance(instrument, QbloxQRM):
                result = instrument.acquire_result()
                if result is not None:
                    results.append(result)

        if len(results) > 1:
            raise ValueError(
                f"Acquisition from multiple instruments is not supported. Obtained a total of {len(results)} results."
            )

        if len(results) == 0:
            raise AttributeError(f"The bus {self.alias} cannot acquire results.")

        return results[0]

    def acquire_qprogram_results(
        self, acquisitions: dict[str, AcquisitionData], channel_id: ChannelID | None = None
    ) -> list[MeasurementResult]:
        """Read the result from the instruments

        Returns:
            list[Result]: Acquired results in chronological order
            channel_id (int, optional): instrument channel of QRM. Defaults to None.
        """
        # TODO: Support acquisition from multiple instruments
        total_results: list[list[MeasurementResult]] = []
        for instrument in self.instruments:
            if isinstance(instrument, QbloxQRM):
                instrument_results = instrument.acquire_qprogram_results(
                    acquisitions=acquisitions, channel_id=channel_id
                )
                total_results.append(instrument_results)

        if len(total_results) == 0:
            raise AttributeError(
                f"The bus {self.alias} cannot acquire results because it doesn't have a readout system control."
            )

        return total_results[0]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FactoryElement class"""

from qililab.typings.enums import (
    ConnectionName,
    InstrumentControllerName,
    InstrumentName,
    PulseDistortionName,
    PulseShapeName,
    ResultName,
)


class FactoryElement:
    """Class FactoryElement"""

    name: PulseDistortionName | PulseShapeName | ResultName | InstrumentName | ConnectionName | InstrumentControllerName

    def __hash__(self) -> int:
        return hash(repr(self))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Rectangular pulse shape."""

from dataclasses import dataclass

import numpy as np

from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class Cosine(PulseShape):
    r"""Cosine pulse shape like :math:`A/2 (1-\lambda_1\cos(\phi)-\lambda_2\cos(2\phi))`, giving a modified sinusoidal-gaussian.

    - :math:`\lambda_1` cosine :math:`A/2 (1-\cos(x))`: Starts and ends at height 0, with a maximum height A in the middle. Which is a sinusoidal like gaussian:

        .. image:: /classes_images/cos0.png
                :width: 300
                :align: center

    - :math:`\lambda_2` cosine :math:`A/2 (1-\cos(2x))`: Starts and ends at height 0, having two maximums heights A, with a node in between. Which is two sinusoidal like gaussians:

        .. image:: /classes_images/cos1.png
                :width: 300
                :align: center

    The finally function would be a weigthed sum of both ``lambda_1`` and ``lambda_2`` contributions, giving an intermediate modified sinusoidal-gaussian.
    Check the following graph from Wolframalpha, where y is the lambda_2 parameter: [https://imgur.com/a/tjatZsg].

    Examples:
        To get the envelope of a Cosine shape, with ``lambda_2`` equal to ``X``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import Cosine

            rectangular_envelope = Cosine(lambda_2=X).envelope(amplitude=1.0, duration=50)

        which for ``X`` being ``0.2`` and ``0.5``, look respectively like:

        .. image:: /classes_images/cosines.png
            :width: 800
            :align: center

    References:
        - Supplement. material B. "Flux pulse parametrization": [https://arxiv.org/abs/1903.02492].
        - OPTIMAL SOLUTION: SMALL CHANGE IN θ: [https://arxiv.org/abs/1402.5467].

    Args:
        lambda_2 (float, optional): Parameter for moving the function :math:`A/2*(1-\cos(x))` into :math:`A/2*(1-\lambda_1\cos(x)-\lambda_2\cos(2x))`
                    which fulfills the constrain: :math:`1=\lambda_1+\lambda_2`. Defaults to 0.
    """

    name = PulseShapeName.COSINE  # noqa: RUF045
    lambda_2: float = 0.0  # between 0 and 1

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0) -> np.ndarray:
        """Modified sinusoidal-gaussian envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """
        x_values = np.linspace(start=0, stop=2 * np.pi, num=int(duration / resolution))
        return amplitude / 2 * (1 - (1 - self.lambda_2) * np.cos(x_values) - self.lambda_2 * np.cos(2 * x_values))



import copy
from unittest.mock import MagicMock, patch

import pytest

from qililab.instrument_controllers.qblox.qblox_cluster_controller import QbloxClusterController
from qililab.instruments.qblox import QbloxQCM
from qililab.platform import Platform
from qililab.data_management import build_platform


@pytest.fixture(name="platform")
def fixture_platform():
    return build_platform(runcard="tests/instrument_controllers/qblox/qblox_runcard.yaml")

class TestQbloxClusterController:
    """Unit tests checking the QDAC-II controller attributes and methods"""

    def test_initialization(self, platform: Platform):
        """Test QDAC-II controller has been initialized correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="cluster_controller")
        assert isinstance(controller_instance, QbloxClusterController)

        controller_settings = controller_instance.settings
        assert isinstance(controller_settings, QbloxClusterController.QbloxClusterControllerSettings)

        controller_modules = controller_instance.modules
        assert len(controller_modules) == 1
        assert isinstance(controller_modules[0], QbloxQCM)

    @patch("qililab.instrument_controllers.qblox.qblox_cluster_controller.Cluster", autospec=True)
    def test_initialize_device(self, device_mock: MagicMock, platform: Platform):
        """Test QDAC-II controller initializes device correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="cluster_controller")

        controller_instance._initialize_device()

        device_mock.assert_called_once_with(name=f"{controller_instance.name.value}_{controller_instance.alias}", identifier=controller_instance.address)

    @patch("qililab.instrument_controllers.qblox.qblox_cluster_controller.Cluster", autospec=True)
    def test_initial_setup(self, device_mock: MagicMock, platform: Platform):
        """Test QDAC-II controller initializes device correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="cluster_controller")

        controller_instance.connect()

        controller_instance.device.reference_source = MagicMock()
        controller_instance.initial_setup()

        controller_instance.device.reference_source.assert_called_once_with(controller_instance.reference_clock.value)

    @patch("qililab.instrument_controllers.qblox.qblox_cluster_controller.Cluster", autospec=True)
    def test_reset(self, device_mock: MagicMock, platform: Platform):
        """Test QDAC-II controller initializes device correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="cluster_controller")
        for module in controller_instance.modules:
            module.clear_cache = MagicMock()

        controller_instance.connect()
        controller_instance.reset()

        controller_instance.device.reset.assert_called_once()
        for module in controller_instance.modules:
            module.clear_cache.assert_called_once()

    def test_check_supported_modules_raises_exception(self, platform: Platform):
        """Test QDAC-II controller raises an error if initialized with wrong module."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias="cluster_controller_wrong_module")
        with pytest.raises(ValueError):
            controller_instance._check_supported_modules()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Scope Acquisitions Result"""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt

from qililab.result.acquisition import Acquisition
from qililab.result.acquisitions import Acquisitions
from qililab.result.qblox_results.scope_data import ScopeData
from qililab.utils.signal_processing import demodulate

from .constants import SCOPE_ACQ_MAX_DURATION


@dataclass
class QbloxScopeAcquisitions(Acquisitions):
    """Qblox Scope Acquisitions Result
    Args:
        scope: ScopeData
        pulse_length (int): Duration (in ns) of the pulse

    """

    scope: ScopeData
    pulse_length: int

    def __post_init__(self):
        """Create acquisitions"""
        i_values, q_values = self._iq_values()
        self._acquisitions = [Acquisition(integration_length=self.pulse_length, i_values=i_values, q_values=q_values)]
        self.data_dataframe_indices = set().union(*[acq.data_dataframe_indices for acq in self._acquisitions])

    def demodulated(self, frequency: float, phase_offset: float = 0.0) -> QbloxScopeAcquisitions:
        """Returns a demodulated QbloxScopeAcquisitions object.

        Args:
            frequency (float): demodulation frequency
            phase_offset (float, optional): demodulation phase offset. Defaults to 0.0.

        Returns:
            QbloxScopeAcquisitions: demodulated QbloxScopeAcquisitions object.
        """
        i_values, q_values = self._iq_values()
        i_demod, q_demod = demodulate(i=i_values, q=q_values, frequency=frequency, phase_offset=phase_offset)
        common_out_of_range = self.scope.path0.out_of_range or self.scope.path1.out_of_range
        i_demod_pathdata = ScopeData.PathData(
            data=i_demod.tolist(), avg_cnt=self.scope.path0.avg_cnt, out_of_range=common_out_of_range
        )
        q_demod_pathdata = ScopeData.PathData(
            data=q_demod.tolist(), avg_cnt=self.scope.path1.avg_cnt, out_of_range=common_out_of_range
        )
        demod_scope_data = ScopeData(path0=i_demod_pathdata, path1=q_demod_pathdata)
        return QbloxScopeAcquisitions(scope=demod_scope_data, pulse_length=self.pulse_length)

    def integrated(self, integrate_from: int = 0, integrate_to: int = SCOPE_ACQ_MAX_DURATION) -> QbloxScopeAcquisitions:
        """Returns an integrated QbloxScopeAcquisitions object.

        Args:
            integrate_from (int, optional): index of the beginning of the integration. Defaults to 0.
            integrate_to (int, optional): index of the end of the integration. Defaults to SCOPE_ACQ_MAX_DURATION.

        Raises:
            IndexError: Integration range is out of bounds.
            IndexError: 'integrate_from' must be smaller than 'integrate_to'.

        Returns:
            QbloxScopeAcquisitions: integrated QbloxScopeAcquisitions object.
        """
        i_values, q_values = self._iq_values()
        if len(i_values) <= integrate_from or len(i_values) < integrate_to:
            raise IndexError("Integration range is out of bounds.")
        if integrate_from >= integrate_to:
            raise IndexError("'integrate_from' must be smaller than 'integrate_to'.")
        integrated_i = np.mean(i_values[integrate_from:integrate_to])
        integrated_q = np.mean(q_values[integrate_from:integrate_to])
        integrated_acquisitions = deepcopy(self)
        integrated_acquisitions.scope.path0.data = [integrated_i]
        integrated_acquisitions.scope.path1.data = [integrated_q]
        return integrated_acquisitions

    def _iq_values(self) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
        """Arrays of IQ values

        Returns:
            tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]: _description_
        """
        i_values = np.array(self.scope.path0.data, dtype=np.float32)
        q_values = np.array(self.scope.path1.data, dtype=np.float32)
        return i_values, q_values



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox pulsar QRM class"""

from dataclasses import dataclass
from typing import Sequence, cast

from qililab.config import logger
from qililab.instruments.decorators import check_device_initialized
from qililab.instruments.qblox.qblox_adc_sequencer import QbloxADCSequencer
from qililab.instruments.qblox.qblox_module import QbloxModule
from qililab.instruments.utils import InstrumentFactory
from qililab.qprogram.qblox_compiler import AcquisitionData
from qililab.result.qblox_results import QbloxResult
from qililab.result.qprogram.qblox_measurement_result import QbloxMeasurementResult
from qililab.typings import AcquireTriggerMode, ChannelID, InstrumentName, IntegrationMode, Parameter, ParameterValue


@InstrumentFactory.register
class QbloxQRM(QbloxModule):
    """Qblox QRM class.

    Args:
        settings (QBloxQRMSettings): Settings of the instrument.
    """

    name = InstrumentName.QBLOX_QRM
    _scoping_sequencer: int | None = None

    @dataclass
    class QbloxQRMSettings(QbloxModule.QbloxModuleSettings):
        """Contains the settings of a specific QRM."""

        awg_sequencers: Sequence[QbloxADCSequencer]

        def __post_init__(self):
            """build AWGQbloxADCSequencer"""
            if len(self.awg_sequencers) > QbloxModule._NUM_MAX_SEQUENCERS:
                raise ValueError(
                    f"The number of sequencers must be greater than 0 and less or equal than {QbloxModule._NUM_MAX_SEQUENCERS}. Received: {len(self.awg_sequencers)}"
                )

            self.awg_sequencers = [
                QbloxADCSequencer(**sequencer) if isinstance(sequencer, dict) else sequencer
                for sequencer in self.awg_sequencers
            ]
            super().__post_init__()

    settings: QbloxQRMSettings

    def is_awg(self) -> bool:
        """Returns True if instrument is an AWG."""
        return True

    def is_adc(self) -> bool:
        """Returns True if instrument is an ADC."""
        return True

    @check_device_initialized
    def initial_setup(self):
        """Initial setup"""
        super().initial_setup()
        self._obtain_scope_sequencer()
        for sequencer in self.awg_sequencers:
            sequencer_id = sequencer.identifier
            # Remove all acquisition data
            self.device.delete_acquisition_data(sequencer=sequencer_id, all=True)
            self._set_integration_length(
                value=cast("QbloxADCSequencer", sequencer).integration_length, sequencer_id=sequencer_id
            )
            self._set_acquisition_mode(
                value=cast("QbloxADCSequencer", sequencer).scope_acquire_trigger_mode, sequencer_id=sequencer_id
            )
            self._set_scope_hardware_averaging(
                value=cast("QbloxADCSequencer", sequencer).scope_hardware_averaging, sequencer_id=sequencer_id
            )
            self._set_hardware_demodulation(
                value=cast("QbloxADCSequencer", sequencer).hardware_demodulation, sequencer_id=sequencer_id
            )
            self._set_threshold(value=cast("QbloxADCSequencer", sequencer).threshold, sequencer_id=sequencer_id)
            self._set_threshold_rotation(
                value=cast("QbloxADCSequencer", sequencer).threshold_rotation, sequencer_id=sequencer_id
            )

    def _map_connections(self):
        """Disable all connections and map sequencer paths with output/input channels."""
        # Disable all connections
        self.device.disconnect_outputs()
        self.device.disconnect_inputs()

        for sequencer_dataclass in self.awg_sequencers:
            sequencer = self.device.sequencers[sequencer_dataclass.identifier]
            for path, output in zip(["I", "Q"], sequencer_dataclass.outputs):
                getattr(sequencer, f"connect_out{output}")(path)

            sequencer.connect_acq_I("in0")
            sequencer.connect_acq_Q("in1")

    def _obtain_scope_sequencer(self):
        """Checks that only one sequencer is storing the scope and saves that sequencer in `_scoping_sequencer`

        Raises:
            ValueError: The scope can only be stores in one sequencer at a time.
        """
        for sequencer in self.awg_sequencers:
            if sequencer.scope_store_enabled:
                if self._scoping_sequencer is None:
                    self._scoping_sequencer = sequencer.identifier
                else:
                    raise ValueError("The scope can only be stored in one sequencer at a time.")

    @check_device_initialized
    def acquire_result(self) -> QbloxResult:
        """Wait for sequencer to finish sequence, wait for acquisition to finish and get the acquisition results.
        If any of the timeouts is reached, a TimeoutError is raised.

        Returns:
            QbloxResult: Class containing the acquisition results.

        """
        results = []
        integration_lengths = []
        for sequencer in self.awg_sequencers:
            if sequencer.identifier in self.sequences:
                sequencer_id = sequencer.identifier
                flags = self.device.get_sequencer_status(
                    sequencer=sequencer_id, timeout=cast("QbloxADCSequencer", sequencer).sequence_timeout
                )
                logger.info("Sequencer[%d] flags: \n%s", sequencer_id, flags)
                self.device.get_acquisition_status(
                    sequencer=sequencer_id, timeout=cast("QbloxADCSequencer", sequencer).acquisition_timeout
                )

                if sequencer.scope_store_enabled:
                    self.device.store_scope_acquisition(sequencer=sequencer_id, name="default")

                for key, data in self.device.get_acquisitions(sequencer=sequencer.identifier).items():
                    acquisitions = data["acquisition"]
                    # parse acquisition index
                    _, qubit, measure = key.split("_")
                    qubit = int(qubit[1:])
                    measurement = int(measure)
                    acquisitions["qubit"] = qubit
                    acquisitions["measurement"] = measurement
                    results.append(acquisitions)
                    integration_lengths.append(sequencer.integration_length)
                self.device.sequencers[sequencer.identifier].sync_en(False)
                integration_lengths.append(sequencer.integration_length)
                self.device.delete_acquisition_data(sequencer=sequencer_id, all=True)

        return QbloxResult(integration_lengths=integration_lengths, qblox_raw_results=results)

    @check_device_initialized
    def acquire_qprogram_results(
        self, acquisitions: dict[str, AcquisitionData], channel_id: ChannelID
    ) -> list[QbloxMeasurementResult]:
        """Read the result from the AWG instrument

        Args:
            acquisitions (list[str]): A list of acquisitions names.

        Returns:
            list[QbloxQProgramMeasurementResult]: Acquired Qblox results in chronological order.
        """
        results = []
        for acquisition, acquisition_data in acquisitions.items():
            sequencer = next(
                (sequencer for sequencer in self.awg_sequencers if sequencer.identifier == channel_id), None
            )
            if sequencer is not None and sequencer.identifier in self.sequences:
                self.device.get_acquisition_status(
                    sequencer=sequencer.identifier,
                    timeout=cast("QbloxADCSequencer", sequencer).acquisition_timeout,
                )
                if acquisition_data.save_adc:
                    self.device.store_scope_acquisition(sequencer=sequencer.identifier, name=acquisition)
                raw_measurement_data = self.device.get_acquisitions(sequencer=sequencer.identifier)[acquisition][
                    "acquisition"
                ]
                measurement_result = QbloxMeasurementResult(
                    bus=acquisitions[acquisition].bus,
                    raw_measurement_data=raw_measurement_data,
                    shape=acquisition_data.shape,
                )
                results.append(measurement_result)

                # always deleting acquisitions without checking save_adc flag
                self.device.delete_acquisition_data(sequencer=sequencer.identifier, name=acquisition)
        return results

    def _set_device_hardware_demodulation(self, value: bool, sequencer_id: int):
        """set hardware demodulation

        Args:
            value (bool): value to update
            sequencer_id (int): sequencer to update the value
        """
        self.device.sequencers[sequencer_id].demod_en_acq(value)

    def _set_device_acquisition_mode(self, mode: AcquireTriggerMode, sequencer_id: int):
        """set acquisition_mode for the specific channel

        Args:
            mode (AcquireTriggerMode): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """
        if sequencer_id != self._scoping_sequencer:
            return
        self.device.scope_acq_sequencer_select(sequencer_id)
        self.device.scope_acq_trigger_mode_path0(mode.value)
        self.device.scope_acq_trigger_mode_path1(mode.value)

    def _set_device_integration_length(self, value: int, sequencer_id: int):
        """set integration_length for the specific channel

        Args:
            value (int): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        self.device.sequencers[sequencer_id].integration_length_acq(value)

    def _set_device_scope_hardware_averaging(self, value: bool, sequencer_id: int):
        """set scope_hardware_averaging for the specific channel

        Args:
            value (bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        if sequencer_id != self._scoping_sequencer:
            return
        self.device.scope_acq_sequencer_select(sequencer_id)
        self.device.scope_acq_avg_mode_en_path0(value)
        self.device.scope_acq_avg_mode_en_path1(value)

    def _set_device_threshold(self, value: float, sequencer_id: int):
        """Sets the threshold for classification at the specific channel.

        Args:
            value (float): integrated value of the threshold
            sequencer_id (int): sequencer to update the value
        """
        integrated_value = value * self.device.sequencers[sequencer_id].integration_length_acq()
        self.device.sequencers[sequencer_id].thresholded_acq_threshold(integrated_value)

    def _set_device_threshold_rotation(self, value: float, sequencer_id: int):
        """Sets the threshold rotation for classification at the specific channel.

        Args:
            value (float): threshold rotation value in degrees (0.0 to 360.0).
            sequencer_id (int): sequencer to update the value
        """
        self.device.sequencers[sequencer_id].thresholded_acq_rotation(value)

    def _set_nco(self, sequencer_id: int):
        """Enable modulation/demodulation of pulses and setup NCO frequency."""
        super()._set_nco(sequencer_id=sequencer_id)
        if cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).hardware_demodulation:
            self._set_hardware_demodulation(
                value=self.get_sequencer(sequencer_id).hardware_modulation, sequencer_id=sequencer_id
            )

    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """set a specific parameter to the instrument"""
        if channel_id is None:
            raise ValueError("channel not specified to update instrument")

        channel_id = int(channel_id)
        if parameter == Parameter.SCOPE_HARDWARE_AVERAGING:
            self._set_scope_hardware_averaging(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.HARDWARE_DEMODULATION:
            self._set_hardware_demodulation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SCOPE_ACQUIRE_TRIGGER_MODE:
            self._set_acquisition_mode(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.INTEGRATION_LENGTH:
            self._set_integration_length(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SAMPLING_RATE:
            self._set_sampling_rate(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.INTEGRATION_MODE:
            self._set_integration_mode(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SEQUENCE_TIMEOUT:
            self._set_sequence_timeout(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.ACQUISITION_TIMEOUT:
            self._set_acquisition_timeout(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.SCOPE_STORE_ENABLED:
            self._set_scope_store_enabled(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.THRESHOLD:
            self._set_threshold(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.THRESHOLD_ROTATION:
            self._set_threshold_rotation(value=value, sequencer_id=channel_id)
            return
        if parameter == Parameter.TIME_OF_FLIGHT:
            self._set_time_of_flight(value=value, sequencer_id=channel_id)
            return
        super().set_parameter(parameter=parameter, value=value, channel_id=channel_id)

    def _set_scope_hardware_averaging(self, value: float | str | bool, sequencer_id: int):
        """set scope_hardware_averaging for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).scope_hardware_averaging = bool(value)

        if self.is_device_active():
            self._set_device_scope_hardware_averaging(value=bool(value), sequencer_id=sequencer_id)

    def _set_threshold(self, value: float | str | bool, sequencer_id: int):
        """Set threshold value for the specific channel.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).threshold = float(value)

        if self.is_device_active():
            self._set_device_threshold(value=float(value), sequencer_id=sequencer_id)

    def _set_threshold_rotation(self, value: float | str | bool, sequencer_id: int):
        """Set threshold rotation value for the specific channel.

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).threshold_rotation = float(value)
        if self.is_device_active():
            self._set_device_threshold_rotation(value=float(value), sequencer_id=sequencer_id)

    def _set_hardware_demodulation(self, value: float | str | bool, sequencer_id: int):
        """set hardware demodulation

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).hardware_demodulation = bool(value)
        if self.is_device_active():
            self._set_device_hardware_demodulation(value=bool(value), sequencer_id=sequencer_id)

    def _set_acquisition_mode(self, value: float | str | bool | AcquireTriggerMode, sequencer_id: int):
        """set acquisition_mode for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).scope_acquire_trigger_mode = AcquireTriggerMode(
            value
        )
        if self.is_device_active():
            self._set_device_acquisition_mode(mode=AcquireTriggerMode(value), sequencer_id=sequencer_id)

    def _set_integration_length(self, value: int | float | str | bool, sequencer_id: int):
        """set integration_length for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).integration_length = int(value)
        if self.is_device_active():
            self._set_device_integration_length(value=int(value), sequencer_id=sequencer_id)

    def _set_sampling_rate(self, value: int | float | str | bool, sequencer_id: int):
        """set sampling_rate for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).sampling_rate = float(value)

    def _set_integration_mode(self, value: float | str | bool | IntegrationMode, sequencer_id: int):
        """set integration_mode for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not string
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).integration_mode = IntegrationMode(value)

    def _set_sequence_timeout(self, value: int | float | str | bool, sequencer_id: int):
        """set sequence_timeout for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float or int
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).sequence_timeout = int(value)

    def _set_acquisition_timeout(self, value: int | float | str | bool, sequencer_id: int):
        """set acquisition_timeout for the specific channel

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not float or int
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).acquisition_timeout = int(value)

    def _set_scope_store_enabled(self, value: float | str | bool, sequencer_id: int):
        """set scope_store_enable

        Args:
            value (float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).scope_store_enabled = bool(value)

    def _set_time_of_flight(self, value: int | float | str | bool, sequencer_id: int):
        """set time_of_flight

        Args:
            value (int | float | str | bool): value to update
            sequencer_id (int): sequencer to update the value

        Raises:
            ValueError: when value type is not bool
        """
        cast("QbloxADCSequencer", self.get_sequencer(sequencer_id)).time_of_flight = int(value)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""File containing the supported native gates."""

import numpy as np
from qibo import Circuit, gates
from qibo.gates.abstract import ParametrizedGate
from qibo.gates.gates import _Un_

from qililab import digital


class Drag(_Un_):
    r"""Native drag pulse dummy class.
    Inherits from qibo unitary gates class

    The native gate is a drag pulse
    .. math::

        R_{MW}(\theta, \phi) = Z_\phi X_\theta Z_{-\phi}

    Please note that the negative Z rotations is applied first! The circuit drawing of this gate looks like the
    following:

    .. math::

        --|RZ(-phi)|--|RX(theta)|--|RZ(phi)|--

    Together with virtual Z gates, this allows us to perform any single-qubit gate, since any
    such gate can be expressed as a unitary
    .. math::

        U(\theta,\phi,\lambda) = Z_\phi X_\theta Z_\lambda &= R_{MW}(\theta, \phi)Z_{\lambda+\phi} &=
        Z_{\phi+\lambda}R_{MW}(\theta, -\lambda)

    Args:
        q (int): qubit where the gate is applied
        theta (float): theta angle of rotation in radians
        phase (float): phase of  the Drag pulse
        trainable (bool): whether parameters are trainable (set to false)
    """

    def __init__(self, q: int, theta: float, phase: float, trainable: bool = True):
        super().__init__(q, trainable=trainable)
        self.name = "drag"
        self.nparams = 2
        self._theta, self._phi = None, None
        self.init_kwargs = {"theta": theta, "phase": phase, "trainable": trainable}
        self.parameter_names = ["theta", "phase"]
        self.parameters = theta, phase


class Wait(ParametrizedGate):
    """The Wait gate.

    Args:
        q (int): the qubit index.
        t (int): the time to wait (ns)
    """

    def __init__(self, q, t):
        super().__init__(trainable=True)
        self.name = "wait"
        self._controlled_gate = None
        self.target_qubits = (q,)

        self.parameters = t
        self.init_args = [q]
        self.init_kwargs = {"t": t}


class _GateHandler:
    """Class to handle the native gates of the Qibo circuit."""

    @staticmethod
    def normalize_angle(angle: float):
        """Normalizes angle in range [-pi, pi].

        Args:
            angle (float): Normalized angle.
        """
        angle %= 2 * np.pi
        if angle > np.pi:
            angle -= 2 * np.pi
        return angle

    @staticmethod
    def get_circuit_gates_info(circuit_gates: list[gates.Gate]) -> list[tuple]:
        """Get the gates of the circuit.

        Args:
            circuit_gates (list[gates.Gate]): list of native gates of the Qibo circuit.

        Returns:
            list[tuple]: List of gates information in the circuit. Where each gate is a tuple of ('name', 'init_args', 'init_kwargs').
        """
        return [(type(gate).__name__, gate.init_args, gate.init_kwargs) for gate in circuit_gates]

    @staticmethod
    def create_gate(gate_class: str, gate_args: list | int, gate_kwargs: dict) -> gates.Gate:
        """Converts a tuple representation of qibo gate (name, qubits) into a Gate object.

        Args:
            gate_class (str): The class name of the gate. Can be any Qibo or Qililab supported class.
            gate_args (list | int): The qubits the gate acts on.
            gate_kwargs (dict): The kwargs of the gate.

        Returns:
            gates.Gate: The qibo Gate object.
        """
        # Solve Identity gate, argument int issue:
        gate_args = [gate_args] if isinstance(gate_args, int) else gate_args
        gate_type = digital if gate_class in {"Drag", "Wait"} else gates
        return getattr(gate_type, gate_class)(*gate_args, **gate_kwargs)

    @staticmethod
    def create_qibo_gates_from_gates_info(circuit_gates: list[tuple]) -> list[gates.Gate]:
        """Converts a list of gate info (name, qubits) into a list of Qibo gates.

        Args:
            circuit_gates (list[tuple]): List of gates in the circuit. Where each gate is a tuple of ('name', 'init_args', 'init_kwargs')
            nqubits (int): Number of qubits in the circuit.

        Returns:
            list[gates.Gate]: Gate list of the qibo Circuit.
        """
        # Create optimized circuit, from the obtained non-cancelled list:
        output_gates = []
        for gate, gate_args, gate_kwargs in circuit_gates:
            qibo_gate = _GateHandler.create_gate(gate, gate_args, gate_kwargs)
            output_gates.append(qibo_gate)

        return output_gates

    @staticmethod
    def create_circuit_from_gates(queue: list[gates.Gate], nqubits: int, wire_names: list[int] | None) -> Circuit:
        """
        Creates a quantum circuit from a list of gate operations.

        Args:
            queue (list[gates.Gate]): A list of gate operations to be added to the circuit.
            nqubits (int): The number of qubits in the circuit.
            wire_names (list[int], optional): Wire names of the circuit to create.

        Returns:
            Circuit: The constructed quantum circuit with the specified gates and number of qubits.
        """
        optimized_circuit = Circuit(nqubits)
        optimized_circuit.add(queue)
        optimized_circuit.wire_names = wire_names

        return optimized_circuit

    @staticmethod
    def extract_qubits_from_gate_args(gate_args: list | int) -> list:
        """Extract qubits from gate_args.

        Args:
            gate_args (list | int): The arguments of the gate.

        Returns:
            list: The qubits of the gate in an iterable.
        """
        # Assuming qubits are the first one or two args:
        if isinstance(gate_args, int):
            return [gate_args]
        return gate_args if len(gate_args) <= 2 else gate_args[:2]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MeasurementResult class."""

from abc import ABC, abstractmethod

import numpy as np

from qililab.typings.enums import ResultName


class MeasurementResult(ABC):
    """Result of a single measurement of QProgram."""

    name: ResultName

    def __init__(self, bus: str, shape: tuple | None = None):
        self.bus: str = bus
        self.shape = shape

    @property
    @abstractmethod
    def array(self) -> np.ndarray:
        """Returns the results in a numpy array format.

        Returns:
            np.ndarray: Numpy array containing the results.
        """

    @property
    @abstractmethod
    def threshold(self) -> np.ndarray:
        """Returns the thresholded data for the result.

        Returns:
            np.ndarray: Thresholded data for the result.
        """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nested Dictionary Iterator"""

from typing import Generator

import pandas as pd


def nested_dict_to_path_tuples(dict_obj: dict) -> Generator:
    """This function accepts a nested dictionary as argument and iterate over all values of nested dictionaries and
    lists, returning a tuple where each element contains the full path until arriving to the value.
    """

    for key, value in dict_obj.items():
        if isinstance(value, dict):
            # If value is dict then iterate over all its values
            for pair in nested_dict_to_path_tuples(value):
                yield key, *pair

        elif isinstance(value, (list, tuple)):
            # If value is list then iterate over its enumeration, adding the indices as keys to the dict
            for pair in ((index, value) for (index, value) in enumerate(value)):
                yield key, *pair

        else:
            # If value is not dict type then yield the value
            yield key, value


def nested_dict_to_path_values(dict_obj: dict) -> list:
    """Transform a nested dict into a list of [key0, keyN, value] into a list using the nested_dict_to_path_tuples
    generator"""
    return list(nested_dict_to_path_tuples(dict_obj=dict_obj))


def nested_dict_to_pandas_dataframe(dict_obj: dict) -> pd.DataFrame:
    """Transform a nested dict into a pandas dataframe, with a (multi)index as `key0/key1/.../keyN`, and a single
    column `value`. A more classical table structure can be easily recovered by calling the `.reset_index()` method of
    the dataframe instance."""
    path_values: list = nested_dict_to_path_values(dict_obj)

    indices = (unnested_result[:-1] for unnested_result in path_values)
    values = (unnested_result[-1:] for unnested_result in path_values)

    pandas_index = pd.MultiIndex.from_frame(pd.DataFrame(indices))
    return pd.DataFrame(values, index=pandas_index, columns=["value"])



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import numpy as np

from qililab.qprogram import CrosstalkMatrix, FluxVector
from qililab.settings.analog import FluxControlTopology
from qililab.waveforms import Arbitrary


class AnnealingProgram:
    """Class for an Annealing Program. The program should have the format

    .. code-block:: python

        [
            {"qubit_0": {"sigma_x" : 0, "sigma_y" : 1, "sigma_z" : 2},
            "coupler_1_0 : {...},
            },      # time=0ns
            {...},  # time=1ns
        .
        .
        .
        ]


    Args:
        platform (Any): platform
        annealing_program (list[dict[str, dict[str, float]]]): dictionary with the annealing program with the above structure.
    """

    def __init__(
        self,
        flux_to_bus_topology: list[FluxControlTopology],
        annealing_program: list[dict[str, dict[str, float]]],
    ):
        """Init method"""
        self._flux_to_bus_topology = flux_to_bus_topology
        self._annealing_program = annealing_program
        self._transpiled_program = []  # type: list # [anneal_step[chip_element_flux_line,value]]

    def transpile(self, transpiler: Callable):
        """First implementation of a transpiler, pretty basic but good as a first step. Transpiles from ising coefficients to fluxes

        Args:
            transpiler (Callable): Transpiler to use. The transpiler should take 2 values as arguments (delta, epsilon)
            and return 2 values (phix, phiz)
        """

        # iterate over each anneal step and transpile ising to fluxes
        for annealing_step in self._annealing_program:
            transpiled_step = {}
            for chip_element in annealing_step:
                phix, phiz = transpiler(
                    delta=annealing_step[chip_element]["sigma_x"], epsilon=annealing_step[chip_element]["sigma_z"]
                )
                transpiled_step[f"phix_{self._chip_element_to_short(chip_element)}"] = phix
                transpiled_step[f"phiz_{self._chip_element_to_short(chip_element)}"] = phiz

            self._transpiled_program.append(transpiled_step.copy())

    def _chip_element_to_short(self, chip_element: str) -> str:
        """Parse names from algorithm notation (e.g. qubit_0) to runcard notation (e.g. q0)

        Args:
            chip_element[str]: name of the chip element
        Returns
            str: shorthand notation
        """
        split_element = chip_element.split("_")
        return (
            f"{chip_element[0]}{split_element[1]}"
            if chip_element[0] == "q"
            else f"{chip_element[0]}{split_element[1]}_{split_element[2]}"
        )

    def get_waveforms(
        self, crosstalk_matrix: CrosstalkMatrix | None = None, minimum_clock_time: int = 1
    ) -> dict[str, Arbitrary]:
        """Returns a dictionary containing (bus, waveform) for each flux control from the transpiled fluxes. `AnnealingProgram.transpile` should be run first. The waveform is an arbitrary waveform obtained from the transpiled fluxes.

        Args:
            crosstalk_matrix[CrosstalkMatrix]: crosstalk matrix to correct the flux vectors with. This is usually the inverse of the crosstalk matrix
            in the Calibration file obtained from experiments.
            minimum_clock_time [int]: minimum unit of clock time for the awg (in ns). Waveforms should be multiples of this. Defaults to 1, equivalent to 1ns resolution.
        Returns:
            dict[str, Arbitrary]: Dictionary containing the waveform to be sent to each bus, with xtalk corrected
        """

        # Initialize maps for bus to flux and flux to bus translation
        bus_to_flux_map = {}
        for flux_bus in self._flux_to_bus_topology:
            if flux_bus.flux in self._transpiled_program[0]:
                if flux_bus.bus in bus_to_flux_map:
                    raise ValueError(
                        f"More than one flux pointing at bus {flux_bus.bus} in the runcard flux to bus topology"
                    )
                bus_to_flux_map[flux_bus.bus] = flux_bus.flux
        flux_to_bus_map = {v: k for k, v in bus_to_flux_map.items()}

        # add padding to waveforms if duration is not multiple of minimum clock time
        padded_ns = 0
        if len(self._transpiled_program) % minimum_clock_time != 0:
            padded_ns = (
                minimum_clock_time - len(self._transpiled_program) % minimum_clock_time
                if len(self._transpiled_program) % minimum_clock_time != 0
                else 0
            )

        # Initialize annealing waveforms
        annealing_waveforms = {bus: padded_ns * [0.0] for bus in bus_to_flux_map}  # type: ignore[var-annotated]

        # unravel each point of the anneal program to get timewise arrays of waveforms
        for annealing_step in self._transpiled_program:
            bus_flux_dict = FluxVector.from_dict(
                {flux_to_bus_map[flux_line]: value for flux_line, value in annealing_step.items()}
            )
            flux_dict = (
                bus_flux_dict.set_crosstalk(crosstalk_matrix)
                if crosstalk_matrix is not None
                else bus_flux_dict.flux_vector
            )
            for bus, value in flux_dict.items():
                annealing_waveforms[bus].append(value)

        return {key: Arbitrary(np.array(value)) for key, value in annealing_waveforms.items()}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseEvent class."""

from dataclasses import dataclass, field

import numpy as np

from qililab.constants import PULSEEVENT
from qililab.pulse.pulse import Pulse
from qililab.pulse.pulse_distortion import PulseDistortion
from qililab.utils import Waveforms
from qililab.utils.signal_processing import modulate


@dataclass
class PulseEvent:
    """Object representing a :class:`Pulse` starting at a certain time, and with its corresponding :class:`PulseDistortion`'s.

    This class will receive a :class:`Pulse` object, a start time for it, a list of the :class:`PulseDistortion`'s that such
    pulse will get applied and a qubit alias to where such pulse will be sent.
    It provides functionality for instrument to generate the waveforms and modulation of pulses.

    Args:
        pulse (Pulse): :class:`Pulse` contained in the PulseEvent.
        start_time (int): Start time of the pulse. Value in ns.
        pulse_distortion (list[PulseDistortion]: List of :class:`PulseDistortion` applied to the pulse.
        qubit (int, optional): the qubit alias. Defaults to None.
    """

    pulse: Pulse
    start_time: int
    pulse_distortions: list[PulseDistortion] = field(default_factory=list)
    qubit: int | None = None

    @property
    def duration(self) -> int:
        """Duration of the pulse in ns."""
        return self.pulse.duration

    @property
    def end_time(self) -> int:
        """End time of the pulse in ns."""
        return self.start_time + self.duration

    @property
    def frequency(self) -> float:
        """Frequency of the pulse in Hz."""
        return self.pulse.frequency

    @property
    def phase(self) -> float:
        """Phase of the pulse."""
        return self.pulse.phase

    def modulated_waveforms(self, resolution: float = 1.0) -> Waveforms:
        """Applies digital quadrature amplitude modulation (QAM) to the envelope.

        Args:
            resolution (float, optional): The resolution of the pulse in ns. Defaults to 1.0.

        Returns:
            Waveforms: I and Q modulated waveforms.
        """
        envelope = self.envelope(resolution=resolution)
        i = np.real(envelope)
        q = np.imag(envelope)

        # Convert pulse relative phase to absolute phase by adding the absolute phase at t=start_time.
        phase_offset = self.phase + 2 * np.pi * self.frequency * self.start_time * 1e-9
        imod, qmod = modulate(i=i, q=q, frequency=self.frequency, phase_offset=phase_offset)

        return Waveforms(i=imod.tolist(), q=qmod.tolist())

    def envelope(self, amplitude: float | None = None, resolution: float = 1.0) -> np.ndarray:
        """Returns the pulse envelope with the corresponding distortions applied.

        Args:
            amplitude (float, optional): Amplitude of the envelope. Defaults to None.
            resolution (float, optional): The resolution of the pulse in ns. Defaults to 1.0.

        Returns:
            np.ndarray: Distorted envelope of the pulse.
        """
        envelope = self.pulse.envelope(amplitude=amplitude, resolution=resolution)

        for distortion in self.pulse_distortions:
            envelope = distortion.apply(envelope)

        return envelope

    @classmethod
    def from_dict(cls, dictionary: dict) -> "PulseEvent":
        """Loads PulseEvent object from dictionary.

        Args:
            dictionary (dict): Dictionary representation of the PulseEvent object.

        Returns:
            PulseEvent: PulseEvent Loaded class.
        """
        local_dictionary = dictionary.copy()
        pulse_settings = local_dictionary[PULSEEVENT.PULSE]
        local_dictionary[PULSEEVENT.PULSE] = Pulse.from_dict(pulse_settings)

        pulse_distortions: list[PulseDistortion] = []

        if PULSEEVENT.PULSE_DISTORTIONS in local_dictionary:
            pulse_distortions.extend(
                PulseDistortion.from_dict(pulse_distortion_dict)
                for pulse_distortion_dict in local_dictionary[PULSEEVENT.PULSE_DISTORTIONS]
            )

        local_dictionary[PULSEEVENT.PULSE_DISTORTIONS] = pulse_distortions

        return cls(**local_dictionary)

    def to_dict(self) -> dict:
        """Returns dictionary of pulse.

        Returns:
            dict: Dictionary describing the pulse event.
        """
        return {
            PULSEEVENT.PULSE: self.pulse.to_dict(),
            PULSEEVENT.START_TIME: self.start_time,
            PULSEEVENT.PULSE_DISTORTIONS: [distortion.to_dict() for distortion in self.pulse_distortions],
            "qubit": self.qubit,
        }

    def __lt__(self, other: "PulseEvent") -> bool:
        """Returns True if and only if self.start_time is less than other.start_time

        Args:
            other (PulseEvent): PulseEvent to compare.

        Returns:
            bool: Comparison evaluation.
        """
        return self.start_time < other.start_time



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import deque
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable

import numpy as np
import qpysequence as QPy
import qpysequence.program as QPyProgram
import qpysequence.program.instructions as QPyInstructions
from qpysequence.constants import INST_MAX_WAIT

from qililab.config import logger
from qililab.qprogram.blocks import Average, Block, ForLoop, InfiniteLoop, Loop, Parallel
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.operations import (
    Acquire,
    Measure,
    Operation,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetMarkers,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Variable
from qililab.waveforms import IQPair, Square, Waveform


@dataclass
class AcquisitionData:
    """Class representing the output information generated by QbloxCompiler for an acquisition."""

    bus: str
    save_adc: bool
    shape: tuple


Sequences = dict[str, QPy.Sequence]
Acquisitions = dict[str, dict[str, AcquisitionData]]


class QbloxCompilationOutput:
    """Class representing the output information generated by QbloxCompiler.

    Attributes:
        sequences (Sequence): A dictionary with the buses participating in the QProgram as keys and the corresponding Sequence as values.
        acquisitions (Acquisitions): A dictionary with the buses participating in the acquisitions as keys and the corresponding Acquisitions as values.
    """

    def __init__(self, qprogram: QProgram, sequences: Sequences, acquisitions: Acquisitions):
        self.qprogram = qprogram
        self.sequences = sequences
        self.acquisitions = acquisitions

    def __iter__(self):
        """Allows the class to be unpacked as a tuple (program, config, measurements)."""
        yield self.sequences
        yield self.acquisitions


class BusCompilationInfo:
    """Class representing the information stored by QbloxCompiler for a bus."""

    def __init__(self) -> None:
        # The generated Sequence
        self.qpy_sequence = QPy.Sequence(
            program=QPy.Program(), waveforms=QPy.Waveforms(), acquisitions=QPy.Acquisitions(), weights=QPy.Weights()
        )

        # Acquisitions information
        self.acquisitions: dict[str, AcquisitionData] = {}

        # Dictionaries to hold mappings useful during compilation.
        self.variable_to_register: dict[Variable, QPyProgram.Register] = {}
        self.waveform_to_index: dict[str, int] = {}
        self.weight_to_index: dict[str, int] = {}
        self.acquisition_to_index: dict[str, int] = {}

        # Create and append the main block to the Sequence's program
        main_block = QPyProgram.Block(name="main")
        self.qpy_sequence._program.append_block(main_block)

        # Stacks to manage block hierarchy during compilation
        self.qpy_block_stack: deque[QPyProgram.Block] = deque([main_block])
        self.qprogram_block_stack: deque[Block] = deque()

        # Counters to help with naming and indexing
        self.next_bin_index = 0
        self.next_acquisition_index = 0
        self.loop_counter = 0
        self.average_counter = 0
        self.square_optimization_counter = 0

        # Syncing durations
        self.static_duration = 0
        self.dynamic_durations: list[Variable] = []
        self.sync_durations: list[QPyProgram.Register] = []

        # Syncing marker. If true, a real-time instruction has been added since the last sync or the beginning of the program.
        self.marked_for_sync = False

        # Time of flight. Defaults to minimum_wait_duration and is updated if times_of_flight parameter is provided during compilation.
        self.time_of_flight = QbloxCompiler.minimum_wait_duration

        # Delay. Defaults 0 delay and is updated if delays parameter is provided within the runcard.
        self.delay = 0

        # Latched Paramter flag
        self.upd_param_instruction_pending: bool = False


class QbloxCompiler:
    """A class for compiling QProgram to QBlox hardware."""

    minimum_wait_duration: int = 4

    def __init__(self) -> None:
        # Handlers to map each operation to a corresponding handler function
        self._handlers: dict[type, Callable] = {
            InfiniteLoop: self._handle_infinite_loop,
            Parallel: self._handle_parallel,
            Average: self._handle_average,
            ForLoop: self._handle_for_loop,
            Loop: self._handle_loop,
            SetFrequency: self._handle_set_frequency,
            SetPhase: self._handle_set_phase,
            ResetPhase: self._handle_reset_phase,
            SetGain: self._handle_set_gain,
            SetOffset: self._handle_set_offset,
            SetMarkers: self._handle_set_markers,
            Wait: self._handle_wait,
            Sync: self._handle_sync,
            Measure: self._handle_measure,
            Acquire: self._handle_acquire,
            Play: self._handle_play,
            Block: self._handle_block,
        }

        self._qprogram: QProgram
        self._buses: dict[str, BusCompilationInfo]
        self._sync_counter: int

    def compile(
        self,
        qprogram: QProgram,
        bus_mapping: dict[str, str] | None = None,
        calibration: Calibration | None = None,
        times_of_flight: dict[str, int] | None = None,
        delays: dict[str, int] | None = None,
        markers: dict[str, str] | None = None,
    ) -> QbloxCompilationOutput:
        """Compile QProgram to qpysequence.Sequence

        Args:
            qprogram (QProgram): The QProgram to be compiled
            bus_mapping (dict[str, str], optional): Optional mapping of bus names. Defaults to None.
            times_of_flight (dict[str, int], optional): Optional time of flight of bus. Defaults to None.

        Returns:
            QbloxCompilationOutput:
        """

        def traverse(block: Block):
            delay_implemented = False
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.append(block)
            for element in block.elements:
                if isinstance(element, Play) and not delay_implemented:
                    for bus in self._buses:
                        if self._buses[bus].delay > 0:
                            self._handle_wait(element=Wait(bus=bus, duration=self._buses[bus].delay), delay=True)
                        elif self._buses[bus].delay < 0:
                            for other_buses in self._buses:
                                if other_buses != bus:
                                    self._handle_wait(
                                        element=Wait(bus=other_buses, duration=-self._buses[bus].delay), delay=True
                                    )
                    delay_implemented = True
                handler = self._handlers.get(type(element))
                if not handler:
                    raise NotImplementedError(f"{element.__class__} is currently not supported in QBlox.")
                appended = handler(element)
                if isinstance(element, Block):
                    traverse(element)
                    if not self._qprogram.qblox.disable_autosync and isinstance(
                        element, (ForLoop, Parallel, Loop, Average)
                    ):
                        self._handle_sync(element=Sync(buses=None), delay=True)
                    if appended:
                        for bus in self._buses:
                            self._buses[bus].qpy_block_stack.pop()
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.pop()

        self._qprogram = qprogram
        if bus_mapping is not None:
            self._qprogram = self._qprogram.with_bus_mapping(bus_mapping=bus_mapping)
        if calibration is not None:
            self._qprogram = self._qprogram.with_calibration(calibration=calibration)
        if self._qprogram.has_calibrated_waveforms_or_weights():
            raise RuntimeError(
                "Cannot compile to hardware-native instructions because QProgram contains named operations that are not mapped. Provide a calibration instance containing all necessary mappings."
            )

        self._sync_counter = 0
        self._buses = self._populate_buses()

        # Pre-processing: Update time of flight
        if times_of_flight is not None:
            for bus in self._buses.keys() & times_of_flight.keys():
                self._buses[bus].time_of_flight = times_of_flight[bus]

        # Pre-processing: Update delay
        if delays is not None:
            for bus in self._buses.keys() & delays.keys():
                self._buses[bus].delay = delays[bus]

        # Pre-processing: Set markers ON/OFF
        for bus in self._buses:
            mask = markers[bus] if markers is not None and bus in markers else "0000"
            self._buses[bus].qpy_sequence._program.blocks[0].append_component(QPyInstructions.SetMrk(int(mask, 2)))
            self._buses[bus].qpy_sequence._program.blocks[0].append_component(QPyInstructions.UpdParam(4))
            self._buses[bus].static_duration += 4

        # Recursive traversal to convert QProgram blocks to Sequence
        traverse(self._qprogram._body)

        # Post-processing: Set all markers OFF, add stop instructions and compile
        for bus in self._buses:
            self._buses[bus].qpy_block_stack[0].append_component(component=QPyInstructions.SetMrk(0))
            self._buses[bus].qpy_block_stack[0].append_component(component=QPyInstructions.UpdParam(4))
            self._buses[bus].qpy_block_stack[0].append_component(component=QPyInstructions.Stop())
            self._buses[bus].static_duration += 4
            self._buses[bus].qpy_sequence._program.compile()

        # Return a dictionary with bus names as keys and the compiled Sequence as values.
        sequences = {bus: bus_info.qpy_sequence for bus, bus_info in self._buses.items()}
        acquisitions = {bus: bus_info.acquisitions for bus, bus_info in self._buses.items()}
        return QbloxCompilationOutput(qprogram=self._qprogram, sequences=sequences, acquisitions=acquisitions)

    def _populate_buses(self):
        """Map each bus in the QProgram to a BusCompilationInfo instance.

        Returns:
            A dictionary where the keys are bus names and the values are BusCompilationInfo objects.
        """

        return {bus: BusCompilationInfo() for bus in self._qprogram.buses}

    def _append_to_waveforms_of_bus(self, bus: str, waveform_I: Waveform, waveform_Q: Waveform | None):
        """Append waveforms to Sequence's Waveforms of the given bus.

        Args:
            bus (str): Name of the bus.
            waveform_I (Waveform): I waveform.
            waveform_Q (Waveform | None): Q waveform.
        """

        def handle_waveform(waveform: Waveform | None, default_length: int = 0):
            _hash = QbloxCompiler._hash_waveform(waveform) if waveform else f"zeros {default_length}"

            if _hash in self._buses[bus].waveform_to_index:
                index = self._buses[bus].waveform_to_index[_hash]
                length = next(
                    len(waveform.data)
                    for waveform in self._buses[bus].qpy_sequence._waveforms._waveforms
                    if waveform.index == index
                )
                return index, length

            envelope = waveform.envelope() if waveform else np.zeros(default_length)
            index = self._buses[bus].qpy_sequence._waveforms.add(envelope)
            self._buses[bus].waveform_to_index[_hash] = index
            return index, len(envelope)

        index_I, length_I = handle_waveform(waveform_I, 0)
        index_Q, _ = handle_waveform(waveform_Q, len(waveform_I.envelope()))
        return index_I, index_Q, length_I

    def _append_to_weights_of_bus(self, bus: str, weights: IQPair):
        def handle_weight(waveform: Waveform):
            _hash = QbloxCompiler._hash_waveform(waveform)

            if _hash in self._buses[bus].weight_to_index:
                index = self._buses[bus].weight_to_index[_hash]
                length = next(
                    len(weight.data)
                    for weight in self._buses[bus].qpy_sequence._weights._weights
                    if weight.index == index
                )
                return index, length

            envelope = waveform.envelope()
            length = len(envelope)
            index = self._buses[bus].qpy_sequence._weights.add(envelope)
            self._buses[bus].weight_to_index[_hash] = index
            return index, length

        index_I, length_I = handle_weight(weights.I)
        index_Q, _ = handle_weight(weights.Q)
        return index_I, index_Q, length_I

    def _handle_parallel(self, element: Parallel):
        if not element.loops:
            raise NotImplementedError("Parallel block should contain loops.")
        if any(isinstance(loop, Loop) for loop in element.loops):
            raise NotImplementedError("Loops with arbitrary numpy arrays are not currently supported for QBlox.")

        loops = []
        iterations = []
        for loop in element.loops:
            operation = QbloxCompiler._get_reference_operation_of_loop(loop=loop, starting_block=element)
            start, step, iters = QbloxCompiler._convert_for_loop_values(loop, operation)  # type: ignore[arg-type]
            loops.append((start, step))
            iterations.append(iters)
        iterations = min(iterations)

        # iterations = min(QbloxCompiler._calculate_iterations(loop.start, loop.stop, loop.step) for loop in element.loops)
        # loops = [(QbloxCompiler._convert_for_loop_values(for_loop=loop, operation=QbloxCompiler._get_reference_operation_of_loop(element)))[:2]) for loop in element.loops]

        for bus in self._buses:
            qpy_loop = QPyProgram.IterativeLoop(
                name=f"loop_{self._buses[bus].loop_counter}", iterations=iterations, loops=loops
            )
            for i, loop in enumerate(element.loops):
                self._buses[bus].variable_to_register[loop.variable] = qpy_loop.loop_registers[i]
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].loop_counter += 1
        return True

    def _handle_average(self, element: Average):
        for bus in self._buses:
            qpy_loop = QPyProgram.Loop(name=f"avg_{self._buses[bus].average_counter}", begin=element.shots)
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].average_counter += 1
        return True

    def _handle_infinite_loop(self, _: InfiniteLoop):
        for bus in self._buses:
            qpy_loop = QPyProgram.InfiniteLoop(name=f"infinite_loop_{self._buses[bus].loop_counter}")
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].loop_counter += 1
        return True

    def _handle_for_loop(self, element: ForLoop):
        operation = QbloxCompiler._get_reference_operation_of_loop(element)
        start, step, iterations = QbloxCompiler._convert_for_loop_values(element, operation)
        for bus in self._buses:
            qpy_loop = QPyProgram.IterativeLoop(
                name=f"loop_{self._buses[bus].loop_counter}", iterations=iterations, loops=[(start, step)]
            )
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].variable_to_register[element.variable] = qpy_loop.loop_registers[0]
            self._buses[bus].loop_counter += 1
        return True

    def _handle_loop(self, _: Loop):
        raise NotImplementedError("Loops with arbitrary numpy arrays are not currently supported for QBlox.")

    def _handle_set_frequency(self, element: SetFrequency):
        convert = QbloxCompiler._convert_value(element)
        frequency = (
            self._buses[element.bus].variable_to_register[element.frequency]
            if isinstance(element.frequency, Variable)
            else convert(element.frequency)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetFreq(frequency=frequency)
        )

        # set the frequency a second time - because of a qblox bug where the first frequency in a hardware loop is false (same is done for the gain).
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetFreq(frequency=frequency)
        )
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_set_phase(self, element: SetPhase):
        convert = QbloxCompiler._convert_value(element)
        phase = (
            self._buses[element.bus].variable_to_register[element.phase]
            if isinstance(element.phase, Variable)
            else convert(element.phase)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.SetPh(phase=phase))
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_reset_phase(self, element: ResetPhase):
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.ResetPh())
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_set_gain(self, element: SetGain):
        convert = QbloxCompiler._convert_value(element)
        gain = (
            self._buses[element.bus].variable_to_register[element.gain]
            if isinstance(element.gain, Variable)
            else convert(element.gain)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgGain(gain_0=gain, gain_1=gain)
        )
        # set the gain a second time - because of a qblox bug where the first gain in a hardware loop is false (same is done for the frequency).
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgGain(gain_0=gain, gain_1=gain)
        )
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_set_offset(self, element: SetOffset):
        convert = QbloxCompiler._convert_value(element)
        offset_0 = (
            self._buses[element.bus].variable_to_register[element.offset_path0]
            if isinstance(element.offset_path0, Variable)
            else convert(element.offset_path0)
        )
        if element.offset_path1 is None:
            offset_1 = offset_0
            logger.warning(
                "Qblox requires an offset for the two paths, the offset of the second path has been set to the same as the first path."
            )
        else:
            offset_1 = (
                self._buses[element.bus].variable_to_register[element.offset_path1]  # type: ignore[index]
                if isinstance(element.offset_path1, Variable)
                else convert(element.offset_path1)
            )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgOffs(offset_0=offset_0, offset_1=offset_1)
        )
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_set_markers(self, element: SetMarkers):
        marker_outputs = int(element.mask, 2)
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetMrk(marker_outputs=marker_outputs)
        )
        self._buses[element.bus].upd_param_instruction_pending = True

    def _handle_wait(self, element: Wait, delay: bool = False):
        duration: QPyProgram.Register | int
        if isinstance(element.duration, Variable):
            duration = self._buses[element.bus].variable_to_register[element.duration]
            self._buses[element.bus].dynamic_durations.append(element.duration)
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Wait(wait_time=duration)
            )
        else:
            convert = QbloxCompiler._convert_value(element)
            duration = convert(element.duration)
            if not delay:
                self._buses[element.bus].static_duration += duration
            # loop over wait instructions if static duration is longer than allowed qblox max wait time of 2**16 -4
            self._handle_add_waits(bus=element.bus, duration=duration)

        self._buses[element.bus].marked_for_sync = True

    def _handle_add_waits(self, bus: str, duration: int):
        """Wait for longer than QBLOX INST_MAX_WAIT by looping over wait instructions

        Args:
            element (Wait): wait element
            duration (int): duration to wait in ns
        """

        if self._buses[bus].upd_param_instruction_pending:
            if 4 < duration < 8:  # you cannot play an update param and then a wait bc both have a minimum of 4
                self._buses[bus].qpy_block_stack[-1].append_component(component=QPyInstructions.UpdParam(duration))
            else:
                self._buses[bus].qpy_block_stack[-1].append_component(component=QPyInstructions.UpdParam(4))
                duration -= 4
                if duration > INST_MAX_WAIT:
                    for _ in range(duration // INST_MAX_WAIT):
                        self._buses[bus].qpy_block_stack[-1].append_component(
                            component=QPyInstructions.Wait(wait_time=INST_MAX_WAIT)
                        )
                if duration >= 4:
                    self._buses[bus].qpy_block_stack[-1].append_component(
                        component=QPyInstructions.Wait(wait_time=duration % INST_MAX_WAIT)
                    )
            self._buses[bus].upd_param_instruction_pending = False

        else:  # no instructions pending
            if duration > INST_MAX_WAIT:
                for _ in range(duration // INST_MAX_WAIT):
                    self._buses[bus].qpy_block_stack[-1].append_component(
                        component=QPyInstructions.Wait(wait_time=INST_MAX_WAIT)
                    )
            self._buses[bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Wait(wait_time=duration % INST_MAX_WAIT)
            )

    def _handle_sync(self, element: Sync, delay: bool = False):
        # Get the buses involved in the sync operation.
        buses = set(element.buses or self._buses)

        # If they are zero or one, return.
        if len(buses) <= 1:
            return

        # If there is no bus marked for sync, return.
        if all(not self._buses[bus].marked_for_sync for bus in buses):
            return

        # Is there any bus that has dynamic durations?
        if any(bus for bus in buses if self._buses[bus].dynamic_durations or self._buses[bus].sync_durations):
            # If yes, we must add a sync block that calculates the difference between buses dynamically.
            # But the following doesn't work unfortunetely, so raise an error for now.
            self.__handle_dynamic_sync(buses=buses)
        else:
            # If no, calculating the difference is trivial.
            self.__handle_static_sync(buses=buses, delay=delay)

        # In any case, mark al buses as synced.
        for bus in buses:
            self._buses[bus].marked_for_sync = False

    def __handle_static_sync(self, buses: set[str], delay: bool = False):
        max_duration = max(self._buses[bus].static_duration for bus in buses)
        if delay:
            max_delay = max(self._buses[bus].delay for bus in buses)
        for bus in buses:
            if delay:
                delay_diff = max_delay - self._buses[bus].delay
                duration_diff = max_duration - self._buses[bus].static_duration + delay_diff
            else:
                duration_diff = max_duration - self._buses[bus].static_duration
            if duration_diff > 0:
                # loop over wait instructions if static duration is longer than allowed qblox max wait time of 2**16 -4
                self._handle_add_waits(bus=bus, duration=duration_diff)
                self._buses[bus].static_duration += duration_diff

    def __handle_dynamic_sync(self, buses: set[str]):
        raise NotImplementedError("Dynamic syncing is not implemented yet.")

    def _handle_measure(self, element: Measure):
        """Wrapper for qblox play and acquire methods to be called in a single operation for consistency with QuantumMachines
        measure operation

        Args:
            element (Measure): measure operation
        """
        time_of_flight = self._buses[element.bus].time_of_flight
        play = Play(bus=element.bus, waveform=element.waveform, wait_time=time_of_flight)
        acquire = Acquire(bus=element.bus, weights=element.weights, save_adc=element.save_adc)
        self._handle_play(play)
        self._handle_acquire(acquire)

    def _handle_acquire(self, element: Acquire):
        # TODO: unify with measure when time of flight is implemented
        loops = [
            (i, loop)
            for i, loop in enumerate(self._buses[element.bus].qpy_block_stack)
            if isinstance(loop, QPyProgram.IterativeLoop) and not loop.name.startswith("avg")
        ]
        shape = tuple(loop[1].iterations for loop in loops)
        num_bins = math.prod(loop[1].iterations for loop in loops)
        acquisition_name = f"acquisition_{self._buses[element.bus].next_acquisition_index}"
        self._buses[element.bus].qpy_sequence._acquisitions.add(
            name=acquisition_name,
            num_bins=num_bins,
            index=self._buses[element.bus].next_acquisition_index,
        )
        self._buses[element.bus].acquisitions[acquisition_name] = AcquisitionData(
            bus=element.bus, save_adc=element.save_adc, shape=shape
        )

        index_I, index_Q, integration_length = self._append_to_weights_of_bus(element.bus, weights=element.weights)

        if num_bins == 1:
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.AcquireWeighed(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=self._buses[element.bus].next_bin_index,
                    weight_index_0=index_I,
                    weight_index_1=index_Q,
                    wait_time=integration_length,
                )
            )
        else:
            bin_register = QPyProgram.Register()
            block_index_for_move_instruction = loops[0][0] - 1 if loops else -2
            block_index_for_add_instruction = loops[-1][0] if loops else -1
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=self._buses[element.bus].next_bin_index, register=bin_register),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            register_I, register_Q = QPyProgram.Register(), QPyProgram.Register()
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=index_I, register=register_I),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=index_Q, register=register_Q),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.AcquireWeighed(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=bin_register,
                    weight_index_0=register_I,
                    weight_index_1=register_Q,
                    wait_time=integration_length,
                )
            )
            self._buses[element.bus].qpy_block_stack[block_index_for_add_instruction].append_component(
                component=QPyInstructions.Add(origin=bin_register, var=1, destination=bin_register)
            )
        self._buses[element.bus].static_duration += integration_length
        self._buses[element.bus].next_bin_index = 0  # maybe this counter can be removed completely
        self._buses[element.bus].next_acquisition_index += 1
        self._buses[element.bus].marked_for_sync = True
        self._buses[element.bus].upd_param_instruction_pending = False

    def _handle_play(self, element: Play):
        waveform_I, waveform_Q = element.get_waveforms()
        waveform_variables = element.get_waveform_variables()
        if waveform_variables:
            logger.error("Variables in waveforms are not supported in Qblox.")
            return
        if element.wait_time:
            # The qp.qblox.play() was used. Don't apply optimizations
            index_I, index_Q, _ = self._append_to_waveforms_of_bus(
                bus=element.bus, waveform_I=waveform_I, waveform_Q=waveform_Q
            )
            convert = QbloxCompiler._convert_value(element)
            duration = convert(element.wait_time)
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Play(index_I, index_Q, wait_time=duration)
            )
        elif (
            isinstance(waveform_I, Square)
            and (waveform_Q is None or isinstance(waveform_Q, Square))
            and (waveform_I.duration >= 100)
        ):
            copied_waveform_I: Square = deepcopy(waveform_I)
            copied_waveform_Q: Square | None = deepcopy(waveform_Q)
            duration = copied_waveform_I.duration
            chunk_duration, iterations, remainder = QbloxCompiler.calculate_square_waveform_optimization_values(
                duration
            )
            copied_waveform_I.duration = chunk_duration
            if isinstance(copied_waveform_Q, Square):
                copied_waveform_Q.duration = chunk_duration
            index_I, index_Q, _ = self._append_to_waveforms_of_bus(
                bus=element.bus, waveform_I=copied_waveform_I, waveform_Q=copied_waveform_Q
            )
            loop = QPyProgram.IterativeLoop(
                name=f"square_{self._buses[element.bus].square_optimization_counter}", iterations=iterations
            )
            loop.append_component(component=QPyInstructions.Play(index_I, index_Q, wait_time=chunk_duration))
            self._buses[element.bus].qpy_block_stack[-1].append_component(component=loop)
            if remainder != 0:
                copied_waveform_I.duration = remainder
                if isinstance(copied_waveform_Q, Square):
                    copied_waveform_Q.duration = remainder
                index_I, index_Q, _ = self._append_to_waveforms_of_bus(
                    bus=element.bus, waveform_I=copied_waveform_I, waveform_Q=copied_waveform_Q
                )
                self._buses[element.bus].qpy_block_stack[-1].append_component(
                    component=QPyInstructions.Play(index_I, index_Q, wait_time=remainder)
                )
            self._buses[element.bus].square_optimization_counter += 1
        else:
            index_I, index_Q, duration = self._append_to_waveforms_of_bus(
                bus=element.bus, waveform_I=waveform_I, waveform_Q=waveform_Q
            )
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Play(index_I, index_Q, wait_time=duration)
            )
        self._buses[element.bus].static_duration += duration
        self._buses[element.bus].marked_for_sync = True
        self._buses[element.bus].upd_param_instruction_pending = False

    def _handle_block(self, element: Block):
        pass

    @staticmethod
    def _get_reference_operation_of_loop(loop: Loop | ForLoop, starting_block: Block | None = None):
        def collect_operations(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    yield from collect_operations(element)
                elif any(variable == loop.variable for variable in element.get_variables()):
                    yield element

        starting_block = starting_block or loop
        operations = list(collect_operations(starting_block))

        if not operations:
            return None
        if isinstance(operations[0], Play) and operations[0].get_waveform_variables():
            raise NotImplementedError("TODO: Variables referenced in a loop cannot be used in Play operation.")
        return operations[0]

    @staticmethod
    def _calculate_iterations(start: int | float, stop: int | float, step: int | float):
        if step == 0:
            raise ValueError("Step value cannot be zero")

        # Calculate the raw number of iterations
        raw_iterations = (stop - start + step) / step

        # If the raw number of iterations is very close to an integer, round it to that integer
        # This accounts for potential floating-point inaccuracies
        if abs(raw_iterations - round(raw_iterations)) < 1e-9:
            return round(raw_iterations)

        # Otherwise, if we're incrementing, take the ceiling, and if we're decrementing, take the floor
        return math.floor(raw_iterations) if step > 0 else math.ceil(raw_iterations)

    @staticmethod
    def _convert_for_loop_values(for_loop: ForLoop, operation: Operation):
        convert = QbloxCompiler._convert_value(operation)
        iterations = QbloxCompiler._calculate_iterations(start=for_loop.start, stop=for_loop.stop, step=for_loop.step)
        qblox_start = convert(for_loop.start)
        qblox_stop = convert(for_loop.stop)
        qblox_step = (qblox_stop - qblox_start) // (iterations - 1)
        return (qblox_start, qblox_step, iterations)

    @staticmethod
    def _convert_value(operation: Operation) -> Callable[[Any], int]:
        conversion_map: dict[type[Operation], Callable[[Any], int]] = {
            SetFrequency: lambda x: int(x * 4),
            SetPhase: lambda x: int(x * 1e9 / (2 * np.pi)),
            SetGain: lambda x: int(x * 32_767),
            SetOffset: lambda x: int(x * 32_767),
            Wait: lambda x: int(max(x, QbloxCompiler.minimum_wait_duration)),
            Play: lambda x: int(max(x, QbloxCompiler.minimum_wait_duration)),
        }
        return conversion_map.get(type(operation), lambda x: int(x))

    @staticmethod
    def _hash_waveform(waveform: Waveform):
        hashes = {
            key: (value.__dict__ if isinstance(value, Waveform) else value) for key, value in waveform.__dict__.items()
        }
        return f"{waveform.__class__.__name__} {hashes}"

    @staticmethod
    def calculate_square_waveform_optimization_values(duration):
        def remainder_conditions(chunk_duration):
            remainder = duration % chunk_duration
            return remainder, (chunk_duration >= 4 and (remainder == 0 or remainder >= 4))

        def find_chunk_duration(condition_func):
            for chunk_duration in range(100, 501):
                if chunk_duration <= duration:
                    remainder, valid = remainder_conditions(chunk_duration)
                    if valid and condition_func(remainder):
                        return chunk_duration
            return None

        # First try for remainder == 0
        final_chunk_duration = find_chunk_duration(lambda rem: rem == 0)
        if final_chunk_duration is not None:
            return final_chunk_duration, duration // final_chunk_duration, duration % final_chunk_duration

        # If not found, try for remainder ≥ 4
        final_chunk_duration = find_chunk_duration(lambda rem: rem >= 4)
        if final_chunk_duration is not None:
            return final_chunk_duration, duration // final_chunk_duration, duration % final_chunk_duration

        # If no suitable piece_duration found, fallback to entire duration
        return duration, 1, 0



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Buses class."""

from dataclasses import dataclass

from qililab.platform.components.bus import Bus


@dataclass
class Buses:
    """Class used as a container of :class:`Bus` objects, these are inside the `elements` attribute, as a list.

    You can add more :class:`Bus` objects to the list, you can get the :class:`Bus` object connected to a concrete port
    through the `add()` or `get()` methods respectively.

    And you can also get all the :class:`Bus` objects containing system controls used for readout via the `readout_buses` property.

    Args:
        buses (list[Bus]): List of :class:`Bus` objects.
    """

    elements: list[Bus]

    def add(self, bus: Bus):
        """Add a bus to the list of buses.

        Args:
            bus (Bus): Bus object to append."""
        self.elements.append(bus)

    def get(self, alias: str):
        """Get bus with the given alias.
        Args:
            bus_alias (str): Alias of the bus we want to get.
        """
        return next((bus for bus in self.elements if bus.alias == alias), None)

    def __iter__(self):
        """Redirect __iter__ magic method to iterate over buses."""
        return self.elements.__iter__()

    def __getitem__(self, key):
        """Redirect __get_item__ magic method."""
        return self.elements.__getitem__(key)

    def __len__(self):
        """Redirect __len__ magic method."""
        return len(self.elements)

    def to_dict(self) -> list[dict]:
        """Return a dict representation of the Buses class."""
        return [bus.to_dict() for bus in self.elements]

    def __str__(self) -> str:
        """String representation of the buses

        Returns:
            str: Buses structure representation
        """
        return "\n".join(str(bus) for bus in self.elements)

    @property
    def readout_buses(self) -> list[Bus]:
        """Returns a list of buses containing system controls used for readout."""
        return [bus for bus in self.elements if bus.has_adc()]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Square waveform."""

import numpy as np
from scipy.special import erf

from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.variable import Domain
from qililab.yaml import yaml

from .waveform import Waveform


@yaml.register_class
class FlatTop(Waveform):
    """Flat top Gaussian rise waveform.

    Args:
        duration (int): Duration of the pulse (ns).
        amplitude (float): Maximum amplitude of the pulse.
        gaussian (float, optional): Sigma number of the gaussian pulse shape. Defaults to 0.5.
        buffer (float, optional): Buffer of the waveform. Defaults to 3.0.
    """

    @requires_domain("amplitude", Domain.Voltage)
    @requires_domain("duration", Domain.Time)
    def __init__(self, amplitude: float, duration: int, gaussian: float = 0.5, buffer: float = 3.0):
        super().__init__()
        self.amplitude = amplitude
        self.duration = duration
        self.gaussian = gaussian
        self.buffer = buffer

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Flat top Gaussian rise envelope.

        Args:
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        x = np.arange(0, self.duration, resolution)
        A = self.amplitude
        g = self.gaussian
        buf = self.buffer
        dur = self.duration
        return 0.5 * A * np.real((erf(g * x - buf) - erf(g * (x - (dur + -buf / g)))))

    def get_duration(self) -> int:
        """Get the duration of the waveform.

        Returns:
            int: The duration of the waveform in ns.
        """
        return self.duration



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the needed classes/methods to decompose gates into native gates."""

from collections.abc import Callable

import numpy as np
from qibo import gates

from .native_gates import Drag, Wait


class GateDecompositions:
    """Abstract data structure that holds decompositions of gates."""

    def __init__(self):
        self.decompositions = {}

    def add(self, gate: gates.Gate, decomposition: list[gates.Gate] | Callable[[gates.Gate], list[gates.Gate]]):
        """Register a decomposition for a gate. Note that the decomposition list can have
        non-native gates

        Args:
            gate (gates.Gate): gate to be decomposed
            decomposition (list(gates.Gate)): list of gates for the decomposition

        """
        self.decompositions[gate] = decomposition

    def __call__(self, gate: gates.Gate) -> list[gates.Gate]:
        """Decompose a gate into native gates

        Args:
            gate (gates.Gate): gate to decompose

        Returns:
            list[gates.Gate]: list of gates corresponding to the decomposition of the given gate
        """

        # check that a decomposition exists
        if type(gate) not in self.decompositions:
            raise NotImplementedError(
                f"Gate of type {gate.__class__} is not supported for transpilation. "
                "Supported gates are {self.decompositions.keys()}"
            )

        decomposition = self.decompositions[gate.__class__]
        if callable(decomposition):
            decomposition = decomposition(gate)
        return [g.on_qubits(dict(enumerate(gate.qubits))) for g in decomposition]


def translate_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
    """Maps Qibo gates to a hardware native implementation (CZ, RZ, Drag, Wait and M (Measurement))

    Check public docstring in :meth:`.CircuitTranspiler.gates_to_native()` for more information.

    Args:
        circuit_gates (list[gates.Gate]): list of gates to be decomposed.

    Returns:
        list[gates.Gate]: list of native gates corresponding to input gates
    """

    # define supported gates (native qpu gates + virtual z + measurement)
    supported_gates = (*native_gates(), gates.RZ, gates.M)

    # check which gates are native gates and if not all of them are so, translate
    to_translate = [not isinstance(gate, supported_gates) for gate in circuit_gates]
    new_gates = []

    # If no more gates to translate, finish:
    if sum(to_translate) == 0:
        return circuit_gates

    # iterate through all gates
    for gate, tt in zip(circuit_gates, to_translate):
        if not tt:
            new_gates.append(gate)  # append already native gates
        # distinguish 1 or 2 qubit gates
        else:
            new_gates += qili_dec(gate)
    return translate_gates(new_gates)


def native_gates():
    """List of native hardware gates

    Returns:
        tuple[gates.Gate]: Hardware native gates
    """
    return (Drag, gates.CZ, Wait)


# Mind that the order of the gates is "the inverse" of the operators
# i.e. to perform the operation AB|psi> the order of the operators
# returned as a list must be  [B, A] so that B is applied to |psi> 1st
qili_dec = GateDecompositions()
qili_dec.add(gates.I, [gates.RZ(0, 0)])
qili_dec.add(gates.Align, lambda gate: [Wait(0, gate.parameters[0])])
qili_dec.add(gates.H, [Drag(0, np.pi / 2, -np.pi / 2), gates.RZ(0, np.pi)])
qili_dec.add(gates.X, [Drag(0, np.pi, 0)])
qili_dec.add(
    gates.Y,
    [
        Drag(0, np.pi, 0),
        gates.RZ(0, np.pi),
    ],
)
qili_dec.add(gates.Z, [gates.RZ(0, np.pi)])

qili_dec.add(gates.RX, lambda gate: [Drag(0, gate.parameters[0], 0)])
qili_dec.add(gates.RY, lambda gate: [Drag(0, gate.parameters[0], np.pi / 2)])
qili_dec.add(gates.RZ, lambda gate: [gates.RZ(0, gate.parameters[0] / 2)])
qili_dec.add(gates.U1, lambda gate: [gates.RZ(0, gate.parameters[0])])
qili_dec.add(gates.U2, lambda gate: [gates.U3(0, np.pi / 2, gate.parameters[0], gate.parameters[1])])
qili_dec.add(
    gates.U3,
    lambda gate: [
        Drag(0, gate.parameters[0], -gate.parameters[2] + np.pi / 2),  # qibo's U3 is different from standard U3
        gates.RZ(0, gate.parameters[1] + gate.parameters[2]),
    ],
)
qili_dec.add(gates.S, [gates.RZ(0, np.pi / 2)])
qili_dec.add(gates.SDG, [gates.RZ(0, -np.pi / 2)])
qili_dec.add(gates.T, [gates.RZ(0, np.pi / 4)])
qili_dec.add(gates.TDG, [gates.RZ(0, -np.pi / 4)])

# register CZ decompositions
qili_dec.add(gates.CNOT, [gates.H(1), gates.CZ(0, 1), gates.H(1)])
qili_dec.add(gates.CZ, [gates.CZ(0, 1)])
qili_dec.add(
    gates.SWAP,
    [
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.H(0),
        gates.CZ(1, 0),
        gates.H(0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
qili_dec.add(
    gates.iSWAP,
    [
        gates.U3(0, np.pi / 2.0, 0, -np.pi / 2.0),
        gates.U3(1, np.pi / 2.0, 0, -np.pi / 2.0),
        gates.CZ(0, 1),
        gates.H(0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(0),
        gates.H(1),
    ],
)
qili_dec.add(
    gates.CRX,
    lambda gate: [
        gates.RX(1, gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
    ],
)
qili_dec.add(
    gates.CRY,
    lambda gate: [
        gates.RY(1, gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.RY(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
    ],
)
qili_dec.add(
    gates.CRZ,
    lambda gate: [
        gates.RZ(1, gate.parameters[0] / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
qili_dec.add(
    gates.CU1,
    lambda gate: [
        gates.RZ(0, gate.parameters[0] / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.H(1),
        gates.RZ(1, gate.parameters[0] / 2.0),
    ],
)
qili_dec.add(
    gates.CU2,
    lambda gate: [
        gates.RZ(1, (gate.parameters[1] - gate.parameters[0]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, -np.pi / 4, 0, -(gate.parameters[1] + gate.parameters[0]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, np.pi / 4, gate.parameters[0], 0),
    ],
)
qili_dec.add(
    gates.CU3,
    lambda gate: [
        gates.RZ(1, (gate.parameters[2] - gate.parameters[1]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, -gate.parameters[0] / 2.0, 0, -(gate.parameters[2] + gate.parameters[1]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, gate.parameters[0] / 2.0, gate.parameters[1], 0),
    ],
)
qili_dec.add(
    gates.FSWAP,
    [
        gates.U3(0, np.pi / 2, -np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, np.pi / 2, np.pi / 2),
        gates.CZ(0, 1),
        gates.U3(0, np.pi / 2, 0, -np.pi / 2),
        gates.U3(1, np.pi / 2, 0, np.pi / 2),
        gates.CZ(0, 1),
        gates.U3(0, np.pi / 2, np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, 0, -np.pi),
    ],
)
qili_dec.add(
    gates.RXX,
    lambda gate: [
        gates.H(0),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.H(0),
    ],
)
qili_dec.add(
    gates.RYY,
    lambda gate: [
        gates.RX(0, np.pi / 2),
        gates.U3(1, np.pi / 2, np.pi / 2, -np.pi),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.RX(0, -np.pi / 2),
        gates.U3(1, np.pi / 2, 0, np.pi / 2),
    ],
)
qili_dec.add(
    gates.RZZ,
    lambda gate: [
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
qili_dec.add(
    gates.TOFFOLI,
    [
        gates.CZ(1, 2),
        gates.RX(2, -np.pi / 4),
        gates.CZ(0, 2),
        gates.RX(2, np.pi / 4),
        gates.CZ(1, 2),
        gates.RX(2, -np.pi / 4),
        gates.CZ(0, 2),
        gates.RX(2, np.pi / 4),
        gates.RZ(1, np.pi / 4),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RZ(0, np.pi / 4),
        gates.RX(1, -np.pi / 4),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
